#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The :mod:`relaxit.distributions.GaussianRelaxedBernoulli` contains classes:

- :class:`relaxit.distributions.GaussianRelaxedBernoulli.GaussianRelaxedBernoulli`

"""
from __future__ import print_function

__docformat__ = "restructuredtext"


import torch
from pyro.distributions.torch_distribution import TorchDistribution
from torch.distributions import constraints


class GaussianRelaxedBernoulli(TorchDistribution):
    r"""
    Gaussian-based continuous Relaxed Bernoulli distribution class inheriting from Pyro's TorchDistribution.

    :param loc: Mean of the normal distribution.
    :type loc: torch.Tensor
    :param scale: Standard deviation of the normal distribution.
    :type scale: torch.Tensor
    """

    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.real
    has_rsample = True

    def __init__(
        self, loc: torch.Tensor, scale: torch.Tensor, validate_args: bool = None
    ):
        r"""Initializes the GaussianRelaxedBernoulli distribution.

        :param loc: Mean of the normal distribution.
        :type loc: torch.Tensor
        :param scale: Standard deviation of the normal distribution.
        :type scale: torch.Tensor
        :param validate_args: Whether to validate arguments.
        :type validate_args: bool
        """
        self.loc = loc.float()  # Ensure loc is a float tensor
        self.scale = scale.float()  # Ensure scale is a float tensor
        self.normal = torch.distributions.Normal(0, self.scale)
        super().__init__(validate_args=validate_args)

    @property
    def batch_shape(self) -> torch.Size:
        r"""
        Returns the batch shape of the distribution.

        The batch shape represents the shape of independent distributions.
        For example, if `loc` is vector of length 3,
        the batch shape will be `[3]`, indicating 3 independent Bernoulli distributions.
        """
        return self.loc.shape

    @property
    def event_shape(self) -> torch.Size:
        r"""
        Returns the event shape of the distribution.

        The event shape represents the shape of each individual event.
        """
        return torch.Size()

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        r"""
        Generates a sample from the distribution using the reparameterization trick.

        :param sample_shape: The shape of the sample.
        :type sample_shape: torch.Size
        :return: A sample from the distribution.
        :rtype: torch.Tensor
        """
        eps = self.normal.sample(sample_shape)
        z = torch.clamp(self.loc + eps, 0, 1)
        return z

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        r"""
        Generates a sample from the distribution.

        :param sample_shape: The shape of the sample.
        :type sample_shape: torch.Size
        :return: A sample from the distribution.
        :rtype: torch.Tensor
        """
        with torch.no_grad():
            return self.rsample(sample_shape)

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        r"""
        Computes the log probability of the given value.

        :param value: The value for which to compute the log probability.
        :type value: torch.Tensor
        :return: The log probability of the given value.
        :rtype: torch.Tensor
        """
        if self._validate_args:
            self._validate_sample(value)

        # Compute the log probability using the normal distribution
        log_prob = -((value - self.loc) ** 2) / (2 * self.scale**2) - torch.log(
            self.scale * torch.sqrt(2 * torch.tensor(torch.pi))
        )

        # Adjust for the clipping to [0, 1]
        cdf_0 = torch.distributions.Normal(self.loc, self.scale).cdf(
            torch.zeros_like(value)
        )
        cdf_1 = torch.distributions.Normal(self.loc, self.scale).cdf(
            torch.ones_like(value)
        )
        log_prob = torch.where(value == 0, torch.log(cdf_0), log_prob)
        log_prob = torch.where(value == 1, torch.log(1 - cdf_1), log_prob)

        return log_prob

    def _validate_sample(self, value: torch.Tensor):
        r"""
        Validates the given sample value.

        :param value: The sample value to validate.
        :type value: torch.Tensor
        """
        if self._validate_args:
            if not (value >= 0).all() or not (value <= 1).all():
                raise ValueError("Sample value must be in the range [0, 1]")



from torch.distributions import kl_divergence, register_kl, Normal

from .InvertibleGaussian import InvertibleGaussian


@register_kl(InvertibleGaussian, InvertibleGaussian)
def _kl_igr_igr(p: InvertibleGaussian, q: InvertibleGaussian):
    r"""
    Compute Kullback-Leibler divergence :math:`KL(p \| q)` between two distributions.
    
    Based on the paper https://arxiv.org/abs/1912.09588.

    .. math::

        KL(p \| q) = \int p(x) \log\frac {p(x)} {q(x)} \,dx

    Args:
        p (InvertibleGaussian): A :class:`~relaxit.distributions.InvertibleGaussian` object.
        q (InvertibleGaussian): A :class:`~relaxit.distributions.InvertibleGaussian` object.

    Returns:
        Tensor: A batch of KL divergences of shape `batch_shape`.
    """
    p_normal = Normal(p.loc, p.scale)
    q_normal = Normal(q.loc, q.scale)
    return kl_divergence(p_normal, q_normal)



import torch
import torch.nn.functional as F
from pyro.distributions.torch_distribution import TorchDistribution
from torch.distributions import constraints
from torch.distributions.utils import probs_to_logits, logits_to_probs


class GumbelSoftmaxTopK(TorchDistribution):
    r"""
    Implementation of the Gaussian-Softmax top-K trick from https://arxiv.org/pdf/1903.06059.

    :param a: logits.
    :type a: torch.Tensor
    :param K: how many samples without replacement to pick.
    :type K: torch.Tensor
    :param tau: Temperature hyper-parameter.
    :type tau: torch.Tensor
    :param hard: if `True`, the returned samples will be discretized as one-hot vectors, but will be differentiated as if it is the soft sample in autograd
    :type hard: bool
    :param validate_args: Whether to validate arguments.
    :type validate_args: bool
    """

    arg_constraints = {
        "probs": constraints.unit_interval, 
        "logits": constraints.real,
        "K": constraints.positive_integer,
        "tau": constraints.positive,
    }
    has_rsample = True

    def __init__(
        self,
        probs: torch.Tensor = None,
        logits: torch.Tensor = None,
        K: torch.Tensor = torch.tensor(1),
        tau: torch.Tensor = torch.tensor(0.1),
        hard: bool = True,
        validate_args: bool = None,
    ):
        r"""Initializes the GumbelSoftmaxTopK distribution.

        TODO
        :param K: how many samples without replacement to pick.
        :type K: torch.Tensor
        :param tau: Temperature hyper-parameter.
        :type tau: torch.Tensor
        :param hard: if `True`, the returned samples will be discretized as one-hot vectors, but will be differentiated as if it is the soft sample in autograd
        :type hard: bool
        :param validate_args: Whether to validate arguments.
        :type validate_args: bool
        """
        if (probs is None) == (logits is None):
            raise ValueError("Pass `probs` or `logits`, but not both of them!")
        elif probs is not None:
            self.probs = probs
            self.logits = probs_to_logits(probs)
        else:
            self.logits = logits
            self.probs = logits_to_probs(logits)
        self.K = K.int()  # Ensure K is a int tensor
        self.tau = tau
        self.hard = hard
        super().__init__(validate_args=validate_args)

    @property
    def batch_shape(self) -> torch.Size:
        """
        Returns the batch shape of the distribution.

        The batch shape represents the shape of independent distributions.
        For example, if `loc` is vector of length 3,
        the batch shape will be `[3]`, indicating 3 independent distributions.
        """
        return self.probs.shape

    @property
    def event_shape(self) -> torch.Size:
        """
        Returns the event shape of the distribution.

        The event shape represents the shape of each individual event.
        """
        return torch.Size()

    def rsample(self) -> torch.Tensor:
        """
        Generates a sample from the distribution using the Gaussian-soft max topK trick.

        :return: A sample from the distribution.
        :rtype: torch.Tensor
        """

        top_k_logits = torch.zeros_like(self.probs)
        logits = torch.clone(self.probs)
        for _ in range(self.K):
            top1_gumbel = F.gumbel_softmax(logits, tau=self.tau, hard=self.hard)
            top_k_logits += top1_gumbel
            logits -= top1_gumbel * 1e10  # mask the selected entry

        return top_k_logits

    def sample(self) -> torch.Tensor:
        """
        Generates a sample from the distribution with no grad.

        :return: A sample from the distribution.
        :rtype: torch.Tensor
        """
        with torch.no_grad():
            return self.rsample()

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Computes the log probability of the given value.
        :param value: The value for which to compute the log probability.
        :type value: torch.Tensor
        :return: The log probability of the given value.
        :rtype: torch.Tensor
        """
        if self._validate_args:
            self._validate_sample(value)

        return -F.binary_cross_entropy(self.probs, value, reduction="none")
    
    def _validate_sample(self, value: torch.Tensor):
        """
        Validates the given sample value.

        Args:
        - value (Tensor): The sample value to validate.
        """
        if self._validate_args:
            if self.hard and ((value != 1.0) & (value != 0.0)).any():
                ValueError(
                    f"If `self.hard` is `True`, then all coordinates in `value` must be 0 or 1 and you have {value}"
                )
            if not self.hard and (value < 0).any():
                ValueError(
                    f"If `self.hard` is `False`, then all coordinates in `value` must be >= 0 and you have {value}"
                )



import torch
from pyro.distributions.torch_distribution import TorchDistribution
from torch.distributions import constraints
from torch.distributions.utils import _standard_normal


class InvertibleGaussian(TorchDistribution):
    """
    Invertible Gaussian distribution, as it was presented in https://arxiv.org/abs/1912.09588.

    Parameters:
    - loc (Tensor): The mean (mu) of the normal distribution.
    - scale (Tensor): The standard deviation (sigma) of the normal distribution.
    """

    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.real
    has_rsample = True

    def __init__(self, loc, scale, temperature, validate_args: bool = None):
        """
        Initializes the Invertible Gaussian distribution.

        Args:
        - loc (Tensor): Mean of the normal distribution.
        - scale (Tensor): Standard deviation of the normal distribution.
        - validate_args (bool): Whether to validate arguments.

        The batch shape is inferred from the shape of the parameters (loc and scale),
        meaning it defines how many independent distributions are parameterized.
        """
        self.loc = loc
        self.scale = scale
        self.temperature = temperature
        batch_shape = torch.Size() if loc.dim() == 0 else loc.shape
        super().__init__(batch_shape, validate_args=validate_args)

    @property
    def batch_shape(self):
        """
        Returns the batch shape of the distribution.

        The batch shape represents the shape of independent distributions.
        """
        return self.loc.shape

    @property
    def event_shape(self):
        """
        Returns the event shape of the distribution.

        The event shape represents the shape of each individual event.
        """
        return torch.Size()

    def softmax_plus_plus(self, y, delta=1):
        """
        Compute the softmax++ function.

        Args:
            y (torch.Tensor): Input tensor of shape (batch_size, num_classes).
            tau (float): Temperature parameter.
            delta (float): Additional term delta > 0.

        Returns:
            torch.Tensor: Output tensor of the same shape as y.
        """
        # Scale the input by the temperature
        scaled_y = y / self.temperature

        # Compute the exponentials
        exp_y = torch.exp(scaled_y)

        # Compute the denominator
        denominator = torch.sum(exp_y, dim=-1, keepdim=True) + delta

        # Compute the softmax++
        softmax_pp = exp_y / denominator

        return softmax_pp

    def rsample(self, sample_shape=torch.Size()):
        """
        Generates a sample from the distribution using the reparameterization trick.

        Args:
        - sample_shape (torch.Size): The shape of the generated samples.
        """
        shape = self._extended_shape(sample_shape)
        eps = _standard_normal(shape, dtype=self.loc.dtype, device=self.loc.device)
        y = self.loc + self.scale * eps
        g = self.softmax_plus_plus(y)
        residual = 1 - torch.sum(g, dim=-1, keepdim=True)
        return torch.cat([g, residual], dim=-1)

    def log_prob(self, value):
        """
        Computes the log likelihood of a value.

        Args:
        - value (Tensor): The value for which to compute the log probability.
        """
        if self._validate_args:
            self._validate_sample(value)

        # Separate the value into g and residual
        g = value[..., :-1]
        residual = value[..., -1:]

        # Invert the softmax++ transformation
        y = self.temperature * (torch.log(g) - torch.log(1 - residual))

        # Compute the log probability of the normal distribution
        log_prob_normal = -0.5 * (
            ((y - self.loc) / self.scale) ** 2
            + torch.log(2 * torch.tensor(torch.pi))
            + 2 * torch.log(self.scale)
        )

        # Compute the Jacobian determinant of the softmax++ transformation
        K = g.size(-1) + 1  # Number of classes including the residual
        log_det_jacobian = (
            -(K - 1) * torch.log(self.temperature).item()
            + torch.sum(torch.log(g), dim=-1, keepdim=True)
            + torch.log(residual)
        )

        # Adjust the log probability by the Jacobian determinant
        log_prob = log_prob_normal - log_det_jacobian

        return log_prob

    def _validate_sample(self, value: torch.Tensor):
        """
        Validates the given sample value.

        Args:
        - value (Tensor): The sample value to validate.
        """
        if self._validate_args:
            if not (value >= 0).all() or not (value <= 1).all():
                raise ValueError("Sample value must be in the range [0, 1]")



import torch
from pyro.distributions.torch_distribution import TorchDistribution
from torch.distributions import constraints


class HardConcrete(TorchDistribution):
    """

    Parameters:

    """

    arg_constraints = {
        "alpha": constraints.positive,
        "beta": constraints.positive,
        "xi": constraints.greater_than(1.0),
        "gamma": constraints.less_than(0.0),
    }
    support = constraints.real
    has_rsample = True

    def __init__(
        self,
        alpha: torch.Tensor,
        beta: torch.Tensor,
        xi: torch.Tensor,
        gamma: torch.Tensor,
        validate_args: bool = None,
    ):
        """

        Args:
        - a (Tensor): logits
        - validate_args (bool): Whether to validate arguments.
        """

        self.alpha = alpha.float()  # Ensure a is a float tensor
        self.beta = beta.float()
        self.gamma = gamma.float()
        self.xi = xi.float()

        self.uniform = torch.distributions.Uniform(
            torch.tensor([0.0]).to(alpha.device), torch.tensor([1.0]).to(alpha.device)
        )
        super().__init__(validate_args=validate_args)
        super().__init__(validate_args=validate_args)

    @property
    def batch_shape(self) -> torch.Size:
        """
        Returns the batch shape of the distribution.

        The batch shape represents the shape of independent distributions.
        For example, if `loc` is vector of length 3,
        the batch shape will be `[3]`, indicating 3 independent Bernoulli distributions.
        """
        return self.alpha.shape

    @property
    def event_shape(self) -> torch.Size:
        """
        Returns the event shape of the distribution.

        The event shape represents the shape of each individual event.
        """
        return torch.Size()

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Generates a sample from the distribution using the reparameterization trick.

        Args:
        - sample_shape (torch.Size): The shape of the sample.

        Returns:
        - torch.Tensor: A sample from the distribution.
        """
        u = self.uniform.sample(sample_shape).to(self.alpha.device)
        value = (torch.log(u) - torch.log(1 - u) + torch.log(self.alpha)) / self.beta
        s = torch.nn.functional.sigmoid(value)
        bar_s = s * (self.xi - self.gamma) + self.gamma
        z = torch.clamp(bar_s, 0, 1)
        return z

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Generates a sample from the distribution.

        Args:
        - sample_shape (torch.Size): The shape of the sample.

        Returns:
        - torch.Tensor: A sample from the distribution.
        """
        with torch.no_grad():
            return self.rsample(sample_shape)

    def _q_prob(self, value: torch.Tensor) -> torch.Tensor:
        """prob of q function

        Args:
            value (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        return (
            self.beta
            * self.alpha
            * value ** (-self.beta - 1)
            * (1 - value) ** (-self.beta - 1)
            / (self.alpha * value ** (-self.beta) + (1 - value) ** (-self.beta)) ** 2
        )

    def _Q_prob(self, value: torch.Tensor) -> torch.Tensor:
        """prob of Q function

        Args:
            value (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        return torch.nn.functional.sigmoid(
            self.beta * (torch.log(value) - torch.log(1 - value))
            - torch.log(self.alpha)
        )

    def _q_bar_prob(self, value: torch.Tensor) -> torch.Tensor:
        """prob of q function

        Args:
            value (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        return (
            1
            / torch.abs(self.xi - self.gamma)
            * self._q_prob((value - self.gamma) / (self.xi - self.gamma))
        )

    def _Q_bar_prob(self, value: torch.Tensor) -> torch.Tensor:
        """prob of Q function

        Args:
            value (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        return self._Q_prob((value - self.gamma) / (self.xi - self.gamma))

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Computes the log probability of the given value.

        Args:
        - value (Tensor): The value for which to compute the log probability.

        Returns:
        - torch.Tensor: The log probability of the given value.
        """
        if self._validate_args:
            self._validate_sample(value)

        log_prob = torch.log(
            self._q_bar_prob(value)
        )  ## WTF ??* (self._Q_bar_prob(1) - self._Q_bar_prob(0) ) )
        log_prob = torch.where(value == 0, torch.log(self._Q_bar_prob(value)), log_prob)
        log_prob = torch.where(
            value == 1, torch.log(1 - self._Q_bar_prob(value)), log_prob
        )
        return log_prob

    def _validate_sample(self, value: torch.Tensor):
        """
        Validates the given sample value.

        Args:
        - value (Tensor): The sample value to validate.
        """
        if self._validate_args:
            if not (value >= 0).all() or not (value <= 1).all():
                raise ValueError("Sample value must be in the range [0, 1]")



import torch
from .LogisticNormalSoftmax import LogisticNormalSoftmax
from pyro.distributions import Dirichlet


def lognorm_approximation_fn(
    dirichlet_distribution: Dirichlet,
) -> LogisticNormalSoftmax:
    """
    Approximates a Dirichlet distribution with a LogisticNormalSoftmax distribution.

    Args:
        dirichlet_distribution (Dirichlet): The Dirichlet distribution to approximate.

    Returns:
        LogisticNormalSoftmax: The approximated LogisticNormalSoftmax distribution.
    """
    concentration = dirichlet_distribution.concentration
    num_events = torch.tensor(dirichlet_distribution.event_shape, dtype=torch.float)

    # Compute the location parameter (mu)
    loc = concentration.log() - (1 / num_events) * concentration.log().sum(
        -1
    ).unsqueeze(-1)

    # Compute the scale parameter (sigma)
    scale = 1 / concentration - (1 / num_events) * (
        2 / concentration - (1 / num_events) * (1 / concentration).sum(-1).unsqueeze(-1)
    )

    # Create the LogisticNormalSoftmax distribution
    lognorm_approximation = LogisticNormalSoftmax(loc, scale)

    return lognorm_approximation


def dirichlet_approximation_fn(
    lognorm_distribution: LogisticNormalSoftmax,
) -> Dirichlet:
    """
    Approximates a LogisticNormalSoftmax distribution with a Dirichlet distribution.

    Args:
        lognorm_distribution (LogisticNormalSoftmax): The LogisticNormalSoftmax distribution to approximate.

    Returns:
        Dirichlet: The approximated Dirichlet distribution.
    """
    num_events = torch.tensor(lognorm_distribution.event_shape, dtype=torch.float)
    loc, scale = lognorm_distribution.loc, lognorm_distribution.scale

    # Compute the concentration parameter (alpha)
    concentration = (1 / scale) * (
        1
        - 2 / num_events
        + loc.exp() / (num_events**2) * loc.neg().exp().sum(-1).unsqueeze(-1)
    )

    # Create the Dirichlet distribution
    dirichlet_approximation = Dirichlet(concentration)

    return dirichlet_approximation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The :mod:`relaxit.distributions.CorrelatedRelaxedBernoulli` contains classes:

- :class:`relaxit.distributions.CorrelatedRelaxedBernoulli.CorrelatedRelaxedBernoulli`

"""
from __future__ import print_function

__docformat__ = "restructuredtext"

import torch
from pyro.distributions.torch_distribution import TorchDistribution
from torch.distributions import constraints
from torch.distributions.normal import Normal


class CorrelatedRelaxedBernoulli(TorchDistribution):
    r"""
    Correlated Relaxed Bernoulli distribution class inheriting from Pyro's TorchDistribution.

    :param pi: Selection probability vector.
    :type pi: torch.Tensor
    :param R: Covariance matrix.
    :type R: torch.Tensor
    :param tau: Temperature hyper-parameter.
    :type tau: torch.Tensor
    """

    arg_constraints = {
        "pi": constraints.interval(0, 1),
        "R": constraints.positive_definite,
        "tau": constraints.positive,
    }
    support = constraints.interval(0, 1)
    has_rsample = True

    def __init__(
        self,
        pi: torch.Tensor,
        R: torch.Tensor,
        tau: torch.Tensor,
        validate_args: bool = None,
    ):
        r"""Initializes the CorrelatedRelaxedBernoulli distribution.

        :param pi: Selection probability vector.
        :type pi: torch.Tensor
        :param R: Covariance matrix.
        :type R: torch.Tensor
        :param tau: Temperature hyper-parameter.
        :type tau: torch.Tensor
        :param validate_args: Whether to validate arguments.
        :type validate_args: bool
        """
        if validate_args:
            self._validate_args(pi, R, tau)

        self.pi = pi
        self.R = R
        self.tau = tau
        self.L = torch.linalg.cholesky(
            R
        )  # Cholesky decomposition of the covariance matrix
        super().__init__(validate_args=validate_args)

    @property
    def batch_shape(self) -> torch.Size:
        r"""
        Returns the batch shape of the distribution.

        The batch shape represents the shape of independent distributions.
        For example, if `pi` is a tensor of shape (batch_size, pi_shape),
        the batch shape will be `[batch_size]`, indicating batch_size independent Bernoulli distributions.
        """
        return self.pi.shape[:-1]

    @property
    def event_shape(self) -> torch.Size:
        r"""
        Returns the event shape of the distribution.

        The event shape represents the shape of each individual event.
        For example, if `pi` is a tensor of shape (batch_size, pi_shape),
        the event shape will be `[pi_shape]`.
        """
        return self.pi.shape[-1:]

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        r"""
        Generates a sample from the distribution using the reparameterization trick.

        :param sample_shape: The shape of the sample.
        :type sample_shape: torch.Size
        :return: A sample from the distribution.
        :rtype: torch.Tensor
        """
        # Sample from the standard multivariate normal distribution
        shape = tuple(sample_shape) + tuple(self.pi.shape)
        eps = torch.randn(shape).to(self.pi.device)
        v = torch.einsum("...ij,...j->...i", self.L, eps)

        # Generate correlated uniform random variables
        uk = Normal(0, 1).cdf(v)

        # Generate relaxed multivariate Bernoulli variable
        log_pi = torch.log(self.pi)
        log_one_minus_pi = torch.log(1 - self.pi)
        log_uk = torch.log(uk)
        log_one_minus_uk = torch.log(1 - uk)

        m_tilde = torch.sigmoid(
            (log_pi - log_one_minus_pi + log_uk - log_one_minus_uk) / self.tau
        )

        return m_tilde

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        r"""
        Generates a sample from the distribution.

        :param sample_shape: The shape of the sample.
        :type sample_shape: torch.Size
        :return: A sample from the distribution.
        :rtype: torch.Tensor
        """
        with torch.no_grad():
            return self.rsample(sample_shape)

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        r"""
        Computes the log probability of the given value.

        :param value: The value for which to compute the log probability.
        :type value: torch.Tensor
        :return: The log probability of the given value.
        :rtype: torch.Tensor
        """
        if self._validate_args:
            self._validate_sample(value)

        # Compute the log probability using the normal distribution
        log_prob = Normal(self.pi, self.tau).log_prob(value)

        # Adjust for the clipping to [0, 1]
        cdf_0 = Normal(self.pi, self.tau).cdf(torch.zeros_like(value))
        cdf_1 = Normal(self.pi, self.tau).cdf(torch.ones_like(value))
        log_prob = torch.where(value == 0, torch.log(cdf_0), log_prob)
        log_prob = torch.where(value == 1, torch.log(1 - cdf_1), log_prob)

        return log_prob

    def _validate_sample(self, value: torch.Tensor):
        r"""
        Validates the given sample value.

        :param value: The sample value to validate.
        :type value: torch.Tensor
        """
        if self._validate_args:
            if not (value >= 0).all() or not (value <= 1).all():
                raise ValueError("Sample value must be in the range [0, 1]")



from pyro.distributions import constraints, Normal
from pyro.distributions.torch import TransformedDistribution
from pyro.distributions.transforms import SoftmaxTransform


class LogisticNormalSoftmax(TransformedDistribution):
    r"""
    Creates a logistic-normal distribution parameterized by :attr:`loc` and :attr:`scale`
    that define the base `Normal` distribution transformed with the
    `SoftmaxTransform` such that::

        X ~ LogisticNormal(loc, scale)
        Y = Logistic(X) ~ Normal(loc, scale)

    Args:
        loc (float or Tensor): mean of the base distribution
        scale (float or Tensor): standard deviation of the base distribution
    """

    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.simplex
    has_rsample = True

    def __init__(self, loc, scale, validate_args=None):
        base_dist = Normal(loc, scale, validate_args=validate_args)
        if not base_dist.batch_shape:
            base_dist = base_dist.expand([1])
        super().__init__(base_dist, SoftmaxTransform(), validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(LogisticNormalSoftmax, _instance)
        return super().expand(batch_shape, _instance=new)

    @property
    def loc(self):
        return self.base_dist.base_dist.loc

    @property
    def scale(self):
        return self.base_dist.base_dist.scale



import io
import os
import sys
from setuptools import setup, find_packages

version_info = {}
with open(os.path.join("src", "relaxit", "_version.py")) as f:
    exec(f.read(), version_info)

try:
    long_description = open("README.md", encoding="utf-8").read()
except Exception as e:
    sys.stderr.write("Failed to read README.md: {}\n".format(e))
    sys.stderr.flush()
    long_description = ""

setup(
    name="relaxit",
    version=version_info["__version__"],
    license="MIT",
    author="Daniil Dorin, Igor Ignashin, Nikita Kiselev, Andrey Veprikov",
    author_email="research.n.math@gmail.com",
    description="A Python library for discrete variables relaxation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/intsystems/relaxit",
    package_dir= {"": "src"}, 
    packages=find_packages(where="src"),
    install_requires=["pyro-ppl>=1.9.1"],
)


from .CorrelatedRelaxedBernoulli import CorrelatedRelaxedBernoulli
from .GaussianRelaxedBernoulli import GaussianRelaxedBernoulli
from .GumbelSoftmaxTopK import GumbelSoftmaxTopK
from .HardConcrete import HardConcrete
from .InvertibleGaussian import InvertibleGaussian
from .LogisticNormalSoftmax import LogisticNormalSoftmax
from .StochasticTimesSmooth import StochasticTimesSmooth
from .StraightThroughBernoulli import StraightThroughBernoulli


__all__ = [
    "CorrelatedRelaxedBernoulli",
    "GaussianRelaxedBernoulli",
    "GumbelSoftmaxTopK",
    "HardConcrete",
    "InvertibleGaussian",
    "LogisticNormalSoftmax",
    "StochasticTimesSmooth",
    "StraightThroughBernoulli"
]



import torch
from pyro.distributions import Bernoulli


class StraightThroughBernoulli(Bernoulli):
    r"""
    Implementation of the Straight Through Bernoulli from https://arxiv.org/abs/1910.02176
    
    Creates a Bernoulli distribution parameterized by :attr:`probs`
    or :attr:`logits` (but not both).

    Samples are binary (0 or 1). They take the value `1` with probability `p`
    and `0` with probability `1 - p`.
    
    However, supports gradient flow through parameters due to the 
    straight through gradient estimator.
    """
    has_rsample = True
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def rsample(self, sample_shape: torch.Size = torch.Size()):
        shape = self._extended_shape(sample_shape)
        probs = self.probs.expand(shape)
        sample = torch.bernoulli(probs).detach()
        return sample + probs - probs.detach()